# pylint: disable=too-many-arguments,too-many-instance-attributes
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from collections import namedtuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats

from .statistics import Base_statistic, qccStatistics


class Cusum:
    def __init__(self, data, center=None, std_dev=None, sizes=None, head_start=0, decision_interval=5,
                 se_shift=1, newdata=None):
        data = pd.DataFrame(data)
        if sizes is None:
            sizes = Base_statistic.getSizes(data)
        elif isinstance(sizes, int):
            sizes = np.array([sizes] * len(data))
        elif len(sizes) != len(data):
            raise ValueError('sizes length doesn'' match with data')
        if decision_interval <= 0:
            raise ValueError('decision_interval must be positive')

        if not 0 <= head_start < decision_interval:
            raise ValueError('head_start must be non-negative and less than decision_interval')

        qcc_type = 'xbarone' if any(size == 1 for size in sizes) else 'xbar'
        if len(data.columns) == 1 and any(size > 1 for size in sizes) and std_dev is None:
            raise ValueError('sizes larger than 1 but data appears to be single samples. '
                             'In this case you must provide also the std_dev')

        qccStatistic = qccStatistics.get(qcc_type)
        statistics = qccStatistic.stats(data, sizes)

        if center is None:
            center = statistics.center
        std_dev = qccStatistic.sd(data, std_dev, sizes=sizes)

        self.newdata = None
        if newdata is not None:
            raise NotImplementedError()

        z = (statistics.statistics.flatten() - center) / (std_dev / np.sqrt(sizes))
        ldb = -decision_interval
        udb = decision_interval

        z_f = z - se_shift / 2
        cusum_pos = [max(0, head_start + z_f[0])]
        for zfi in z_f[1:]:
            cusum_pos.append(max(0, cusum_pos[-1] + zfi))
        cusum_pos = np.array(cusum_pos)

        z_f = z + se_shift / 2
        cusum_neg = [max(0, head_start - z_f[0])]
        for zfi in z_f[1:]:
            cusum_neg.append(max(0, cusum_neg[-1] - zfi))
        cusum_neg = np.array([-v for v in cusum_neg])

        violations = {'lower': np.where(cusum_neg < ldb)[0],
                      'upper': np.where(cusum_pos > udb)[0]}

        self.data = data
        self.statistics = statistics.statistics
        self.sizes = sizes
        self.center = center
        self.std_dev = std_dev

        self.pos = cusum_pos
        self.neg = cusum_neg
        self.head_start = head_start
        self.decision_interval = decision_interval
        self.se_shift = se_shift
        self.ldb = ldb
        self.udb = udb
        self.violations = violations

    def plot(self, ax=None, title='cusum Chart', xlabel='Group', ylabel='Cumulative Sum'):
        if ax is None:
            _, ax = plt.subplots(figsize=(8, 6))

        if self.newdata is not None:
            raise NotImplementedError()

        statistics = self.statistics
        indices = list(range(len(statistics)))

        if title is not None:
            ax.set_title(title)

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        ax.axhline(0, linewidth=2, color='grey')
        ax.axhline(self.decision_interval, linestyle=':', color='grey')
        ax.axhline(-self.decision_interval, linestyle=':', color='grey')

        indices = np.array(indices)
        for cs, v in [(self.pos, self.violations['upper']), (self.neg, self.violations['lower'])]:
            v = np.array(v)
            ax.plot(indices, cs, color='grey')
            ax.plot(indices[v], cs[v], marker="s", color='red')
            ax.plot(np.delete(indices, v), np.delete(cs, v), color='black', marker='o', markersize=4)

        ax2 = ax.twinx()
        ax2.set_ylim(*ax.get_ylim())
        ax2.yaxis.tick_right()
        ax2.set_yticks([self.ldb, self.udb])
        ax2.set_yticklabels(['LDB', 'UDB'])

        fig = ax.get_figure()
        fig.subplots_adjust(bottom=0.23)
        fig.text(0.2, 0.1, f'Number of groups = {len(self.statistics)}', fontsize=12)
        fig.text(0.2, 0.06, f'Center = {self.center:.5g}', fontsize=12)
        fig.text(0.2, 0.02, f'StdDev = {self.std_dev:.5g}', fontsize=12)

        fig.text(0.5, 0.1, f'Decision interval (std. err.) = {self.decision_interval}', fontsize=12)
        fig.text(0.5, 0.06, f'Shift detection (std. err.) {self.se_shift}', fontsize=12)
        nrBeyondLimits = len(self.violations['upper']) + len(self.violations['lower'])
        fig.text(0.5, 0.02, f"No. of points beyond limits = {nrBeyondLimits}", fontsize=12)


def cusumArl(*, randFunc=None, N=1000, limit=10_000, seed=None,
             kp=1, km=-1, hp=3, hm=-3,
             side='both', verbose=False):
    side = side.lower()
    if side not in ("both", "upper", "lower"):
        raise ValueError("side = '{side}' is not supported.")
    np.random.seed(seed)

    randFunc = randFunc or stats.norm()

    result = {}
    result['run'] = [runLength(randFunc.rvs(limit), kp, km, hp, hm, side) for _ in range(N)]
    rls = np.array([run.rl for run in result['run']])
    result['rls'] = rls
    # Ignore inf values for statistics calculations
    rls = np.ma.masked_invalid(rls)
    result['statistic'] = {
        'ARL': np.mean(rls),
        'Std. Error': np.sqrt((np.mean(rls ** 2) - np.mean(rls)) / N),
    }
    if verbose:
        statistic = result['statistic']
        print(f"ARL {statistic['ARL']:.5g}  Std. Error {statistic['Std. Error']:.5g}")
    return result


def cusumPfaCed(*, randFunc1=None, randFunc2=None,
                tau=10, N=100, limit=10_000, seed=None,
                kp=1, km=-1, hp=3, hm=-3,
                side='both', verbose=True):
    side = side.lower()
    if side not in ("both", "upper", "lower"):
        raise ValueError("side = '{side}' is not supported.")
    np.random.seed(seed)

    rls = []
    for _ in range(N):
        x = [*randFunc1.rvs(tau), *randFunc2.rvs(limit - tau)]
        rls.append(runLength(x, kp, km, hp, hm, side).rl)
    result = {'rls': rls}

    rls = np.ma.masked_invalid(rls)
    pfa = np.mean(rls < tau)
    ced = np.mean(rls[rls >= tau]) - tau
    se = np.sqrt(
        (np.sum(rls[rls >= tau] ** 2) / (N - np.sum(rls < tau)) - ced ** 2)
        /
        (N - np.sum(rls < tau)))

    result['statistic'] = {
        'PFA': pfa,
        'CED': ced,
        'Std. Error': se
    }
    if verbose:
        statistic = result['statistic']
        print(f"PFA {statistic['PFA']:.5g}  CED {statistic['CED']:.5g}  Std. Error {statistic['Std. Error']:.5g}")
    return result


#
#   res <- list(run = apply(data, MARGIN=1, .runLength, kp=kp, km=km, ubd=hp, lbd=hm, side=side))
#   res$rls <- sapply(res$run, function(x) x$rl)
#
#   pfa <- sum(res$rls < tau)/nrow(data)
#   ced = sum(res$rls[res$rls >= tau])/(nrow(data) - sum(res$rls < tau)) - tau
#   se = sqrt((sum(res$rls[res$rls >= tau]^2)/(nrow(data) - sum(res$rls < tau)) - ced ^ 2) /
#            (nrow(data) - sum(res$rls < tau)))
#
#   res$statistic <- c(pfa, ced, se)
#   names(res$statistic) <- c("PFA", "CED", "Std. Error")
#
#   if(printSummary)
#     print(res$statistic)
#
#   invisible(res)


def runLength(x, kp, km, ubd, lbd, side):
    x = np.array(x)

    if side in ['both', 'upper']:
        cusum = 0
        vUpper = np.inf
        for i, zfi in enumerate(x - kp):
            cusum = max(0, cusum + zfi)
            if cusum > ubd:
                vUpper = i
                break

    if side in ['both', 'lower']:
        cusum = 0
        vLower = np.inf
        for i, zfi in enumerate(x - km):
            cusum = min(0, cusum + zfi)
            if cusum < lbd:
                vLower = i
                break
    Result = namedtuple('RunLength', 'violationsLower,violationsUpper,rl')
    rl = 1
    if side == 'both':
        rl = min(vLower, vUpper)
    elif side == 'upper':
        rl = min(vUpper)
    elif side == 'lower':
        rl = min(vLower)
    return Result(vLower, vUpper, rl)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import numpy as np
import pytest


def assertQCC(qcc, center, std_dev, limits):
    assert qcc.center == pytest.approx(center)
    assert qcc.std_dev == pytest.approx(std_dev)
    if qcc.limits.shape == (1, 2):
        np.testing.assert_allclose(qcc.limits, [limits], rtol=1e-5)
    else:
        np.testing.assert_allclose(qcc.limits[:len(limits)], limits, rtol=1e-5)


def assertQCCviolations(qcc, beyondLimits=None, violatingRuns=None):
    if beyondLimits is None:
        assert len(qcc.violations['beyondLimits']['UCL']) == 0
        assert len(qcc.violations['beyondLimits']['LCL']) == 0
    else:
        np.testing.assert_array_equal(qcc.violations['beyondLimits']['UCL'], beyondLimits['UCL'])
        np.testing.assert_array_equal(qcc.violations['beyondLimits']['LCL'], beyondLimits['LCL'])

    if violatingRuns is None:
        assert len(qcc.violations['violatingRuns']) == 0


def assertNewQCC(qcc, center, std_dev, limits):
    pass



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest

import pytest
from scipy import stats

from mistat.qcc.cusum import cusumArl, cusumPfaCed


class TestCusum(unittest.TestCase):
    def test_cusumArl(self):
        arl = cusumArl(randFunc=stats.norm(), N=100, limit=10000, seed=1, verbose=False)
        assert arl['statistic']['ARL'] == pytest.approx(1013.86)
        assert arl['statistic']['Std. Error'] == pytest.approx(148.36833)

        arl = cusumArl(randFunc=stats.norm(loc=0.5), N=100, limit=10000, seed=1, verbose=False)
        assert arl['statistic']['ARL'] == pytest.approx(101.88)
        assert arl['statistic']['Std. Error'] == pytest.approx(13.180037)

        arl = cusumArl(randFunc=stats.norm(loc=1.0), N=100, limit=10000, seed=1, verbose=False)
        assert arl['statistic']['ARL'] == pytest.approx(16.91)
        assert arl['statistic']['Std. Error'] == pytest.approx(2.171727)

        arl = cusumArl(randFunc=stats.norm(loc=1.5), N=100, limit=10000, seed=1, verbose=False)
        assert arl['statistic']['ARL'] == pytest.approx(5.82)
        assert arl['statistic']['Std. Error'] == pytest.approx(0.6506919)

    def test_cusumArl_inf(self):
        arl = cusumArl(randFunc=stats.binom(n=100, p=0.05), N=100, limit=2000, seed=3,
                       kp=5.95, km=3.92, hp=12.87, hm=-8.66, verbose=False)
        assert arl['statistic']['ARL'] == pytest.approx(414.16161)
        assert arl['statistic']['Std. Error'] == pytest.approx(58.66780)

    def test_cusumPfaCed(self):
        cusumPfaCed(randFunc1=stats.norm(), randFunc2=stats.norm(loc=1),
                    tau=100, N=100, limit=1_000, seed=1, verbose=False)
        # TODO:  write asserts



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import numpy as np
import pandas as pd
from scipy import stats

from .dodge_base import AcceptanceSamplingPlan


# ' Single sampling plans for the binomial, hypergeometric and Poisson
# ' distributions.
def SSPlanBinomial(N, n, Ac, p=None):
    p = p if p is not None else np.arange(0, 0.3, 0.001)
    p = np.array(p)

    OC = stats.binom.cdf(Ac, n, p)

    AOQ = (N - n) * p * OC / N
    ATI = n * OC + N * (1 - OC)
    return AcceptanceSamplingPlan(p=p, OC=OC, n=[n] * len(p), AOQ=AOQ, ATI=ATI, ASN=None)


def SSPlanHyper(N, n, Ac, p=None):
    p = p if p is not None else np.arange(0, 0.3, 0.001)
    p = np.array(p)

    OC = stats.hypergeom.cdf(Ac, N, np.round(N*p), n)

    AOQ = (N - n) * p * OC / N
    ATI = n * OC + N * (1 - OC)
    return AcceptanceSamplingPlan(p=p, OC=OC, n=[n] * len(p), AOQ=AOQ, ATI=ATI, ASN=None)


def SSPlanPoisson(N, n, Ac, p=None):
    p = p if p is not None else np.arange(0, 0.3, 0.001)
    p = np.array(p)

    OC = stats.poisson.cdf(Ac, n*p)
    AOQ = (N - n) * p * OC / N
    ATI = n * OC + N * (1 - OC)
    return AcceptanceSamplingPlan(p=p, OC=OC, n=[n] * len(p), AOQ=AOQ, ATI=ATI, ASN=None)


# ' Single Sampling Plan Designs
# '
# ' Design a single sampling plan for given AQL, alpha, LQL, and beta. Currently
# ' there are functions for the binomial and Poisson distributions.
# '
# ' @param AQL Acceptable quality level
# ' @param alpha producer's risk
# ' @param LQL Limiting quality level
# ' @param beta consumers' risk

def SSPDesignBinomial(AQL, alpha, LQL, beta):
    def nl(Ac, LQL, beta):
        n = 1
        while stats.binom(n, LQL).cdf(Ac) >= beta:
            n += 1
        return n

    def nu(Ac, AQL, alpha):
        n = 1
        while stats.binom(n, AQL).cdf(Ac) >= 1 - alpha:
            n += 1
        return n

    Ac = 0
    while nl(Ac, LQL, beta) > nu(Ac, AQL, alpha):
        Ac += 1
    n = nl(Ac, LQL, beta)
    return pd.Series({'n': n, 'Ac': Ac})


def SSPDesignPoisson(AQL, alpha, LQL, beta):
    def nl(Ac, LQL, beta):
        n = 1
        while stats.poisson(n * LQL).cdf(Ac) >= beta:
            n += 1
        return n

    def nu(Ac, AQL, alpha):
        n = 1
        while stats.poisson(n * AQL).cdf(Ac) >= 1 - alpha:
            n += 1
        return n

    Ac = 0
    while nl(Ac, LQL, beta) > nu(Ac, AQL, alpha):
        Ac += 1
    n = nl(Ac, LQL, beta)
    return pd.Series({'n': n, 'Ac': Ac})



# mypy: disallow_untyped_defs,disallow_untyped_calls
# pylint: disable=too-many-arguments
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from dataclasses import dataclass, field
from enum import Enum
from itertools import product
from typing import List, Optional, Union, cast

import numpy as np
import pandas as pd
from scipy.stats import binom, hypergeom, poisson


class OCtype(Enum):
    binomial = 'binomial'
    hypergeom = 'hypergeom'
    poisson = 'poisson'


IntOrListInt = Union[int, List[int]]
Probabilities = Optional[Union[List[float], np.ndarray]]


def getDistribution(n: IntOrListInt, c: IntOrListInt, distribution: OCtype,
                    r: List[int] = None, pd: Probabilities = None,   # pylint: disable=redefined-outer-name
                    N: int = None, D: List[int] = None) -> 'OCdistribution':
    if not isinstance(n, list):
        n = [n]
    if not isinstance(c, list):
        c = [c]
    if distribution == OCtype.binomial:
        return OCbinomial(n=n, c=c, r=r, pd=pd)
    if distribution == OCtype.hypergeom:
        return OChypergeom(n=n, c=c, r=r, N=N, pd=pd, D=D)
    if distribution == OCtype.poisson:
        return OCpoisson(n=n, c=c, r=r, pd=pd)

    raise NotImplementedError(f'Distribution {distribution} not implemented')


@dataclass
class OCdistribution:
    n: List[int]
    c: List[int]
    r: Optional[List[int]] = None

    distribution: OCtype = field(init=False)
    paccept: Optional[np.ndarray] = None

    def __post_init__(self) -> None:
        if self.r is None:
            if len(self.c) <= 2:
                self.r = [1 + self.c[-1]] * len(self.c)
            else:
                self.r = None

        self.c = np.array(self.c)  # type: ignore
        self.r = np.array(self.r)  # type: ignore


@dataclass
class OCbinomial(OCdistribution):
    pd: Probabilities = None

    def __post_init__(self) -> None:
        super().__post_init__()
        self.distribution = OCtype.binomial
        if self.r is None:
            raise ValueError('r must be provided in class initialization')

        if self.pd is None:
            self.pd = np.linspace(0, 1, 101)

        self.paccept = np.array([self.calcBinomial(pdi, self.n, self.c, self.r) for pdi in self.pd])

    def calcBinomial(self, p_d: float, n: List[int], c: List[int], r: List[int]) -> float:
        # For each stage, find out all the possibilities which could
        # lead to still not having made a decision and then calculate
        # the appropriate probabilities.
        p_acc = 0.0
        for k in range(len(n)):
            # Determine combinations
            comb = [list(range(c[i] + 1, r[i])) for i in range(k)]
            x = pd.DataFrame(list(product(*comb)))
            # Calculate change from previous
            x.iloc[:, 1:] = x.iloc[:, 1:].values - x.iloc[:, :-1].values

            x[k] = c[k] - np.sum(cast(np.ndarray, x), axis=1)
            for _, xi in x.iterrows():
                p_acc += self.probAcc(xi.values, n, p_d)
        return p_acc

    @staticmethod
    def probAcc(x: np.ndarray, n: List[int], p: float) -> float:
        k = len(x) - 1
        f = cast(float, binom.cdf(x[k], n[k], p))
        for i in range(k):
            f *= binom.pmf(x[i], n[i], p)
        return f


@dataclass
class OChypergeom(OCdistribution):
    N: Optional[int] = None
    pd: Probabilities = None
    D: Optional[List[int]] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.N is None:
            raise ValueError('N must be provided in class initialization')
        if self.r is None:
            raise ValueError('r must be provided in class initialization')
        if self.pd is None and self.D is None:
            raise ValueError('One of D or pd must be provided in class initialization')

        self.distribution = OCtype.hypergeom
        if not self.D:
            assert self.pd is not None
            self.D = [round(pdi * self.N) for pdi in self.pd]

        self.paccept = np.array([self.calcHypergeom(Di, self.n, self.c, self.r, self.N) for Di in self.D])

    def calcHypergeom(self, D: int, n: List[int], c: List[int], r: List[int], N: int) -> float:
        # For each stage, find out all the possibilities which could
        # lead to still not having made a decision and then calculate
        # the appropriate probabilities.
        p_acc = 0.0
        for k, c_k in enumerate(c):
            # Determine combinations
            comb = [list(range(c[i] + 1, r[i])) for i in range(k)]
            x = pd.DataFrame(list(product(*comb)))
            # Calculate change from previous
            x.iloc[:, 1:] = x.iloc[:, 1:].values - x.iloc[:, :-1].values

            x[k] = c_k - np.sum(cast(np.ndarray, x), axis=1)
            for _, xi in x.iterrows():
                p_acc += self.probAcc(xi.values, n, N, D)
        return p_acc

    @staticmethod
    def probAcc(x: np.ndarray, n: List[int], N: int, D: int) -> float:
        k = len(x)
        k1 = k - 1
        x_cumsum = np.cumsum(x)[0:k]
        n_cumsum = np.cumsum(n)
        D_cum = D - np.array([0, *x_cumsum[0:k1]])
        N_cum = N - np.array([0, *n_cumsum[0:k1]])
        f = cast(float, hypergeom(N_cum[-1], max(0, D_cum[-1]), n[k1]).cdf(x[-1]))
        for i in range(len(x) - 1):
            f *= hypergeom(N_cum[i], max(0, D_cum[i]), n[i]).pmf(x[i])
        return f


@dataclass
class OCpoisson(OCdistribution):
    pd: Probabilities = None

    def __post_init__(self) -> None:
        super().__post_init__()
        self.distribution = OCtype.poisson
        if self.r is None:
            raise ValueError('r must be provided in class initialization')

        if self.pd is None:
            self.pd = np.linspace(0, 1, 101)

        self.paccept = np.array([self.calcPoisson(pdi, self.n, self.c, self.r) for pdi in self.pd])

    def calcPoisson(self, p_d: float, n: List[int], c: List[int], r: List[int]) -> float:
        # For each stage, find out all the possibilities which could
        # lead to still not having made a decision and then calculate
        # the appropriate probabilities.
        p_acc = 0.0
        for k in range(len(n)):
            # Determine combinations
            comb = [list(range(c[i] + 1, r[i])) for i in range(k)]
            x = pd.DataFrame(list(product(*comb)))
            # Calculate change from previous
            x.iloc[:, 1:] = x.iloc[:, 1:].values - x.iloc[:, :-1].values
            x[k] = c[k] - np.sum(cast(np.ndarray, x), axis=1)
            for _, xi in x.iterrows():
                p_acc += self.probAcc(xi.values, n, p_d)
        return p_acc

    @staticmethod
    def probAcc(x: np.ndarray, n: List[int], p: float) -> float:
        k = len(x) - 1
        f = cast(float, poisson.cdf(x[k], np.floor(n[k] * p + 0.5)))
        for i in range(k):
            f *= poisson.pmf(x[i], n[i] * p)
        return f



import warnings

import statsmodels.formula.api as smf
import statsmodels.stats as sms


def find_best_model_partialF(outcome, variable_sets, data, old_model, opt_max):
    optF = 1 if opt_max else -1
    partialF = 0
    best_model = None
    best_vars = None
    for variables in variable_sets:
        formula = f'{outcome} ~ 1'
        if variables:
            formula = f"{formula} + {' + '.join(variables)}"
        new_model = smf.ols(formula, data=data).fit()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            comparison = sms.anova.anova_lm(old_model, new_model)
        if optF * partialF < optF * comparison.F[1]:
            best_vars = variables
            best_model = new_model
            partialF = comparison.F[1]
    return best_vars, best_model, partialF


def stepwise_regression(outcome, all_vars, data):
    # initial model
    model = smf.ols(f'{outcome} ~ 1', data=data).fit()

    include = set()
    step = 0
    F_to_add = 4
    F_to_remove = 4
    all_vars = set(all_vars)
    while len(include) != len(all_vars):
        step += 1
        partialF = 0

        # try to add variables to model
        candidates = [include | {xl} for xl in all_vars.difference(include)]
        best_vars, best_model, partialF = find_best_model_partialF(outcome,
                                                                   candidates, data, model, True)

        # stop if best partial-F is below cutoff
        if partialF < F_to_add:
            break
        print(f'Step {step} add - (F: {partialF:.2f}) ', end='')
        print(f' {" ".join(sorted(best_vars))}')
        model = best_model
        include = best_vars

        # try to exclude variables from model
        if len(include) < 2:
            continue
        candidates = [include - {xl} for xl in include]
        best_vars, best_model, partialF = find_best_model_partialF(outcome,
                                                                   candidates, data, model, False)

        # continue with adding if best partial-F is above cutoff
        if partialF < F_to_remove:
            continue

        print(f'Step {step} remove - (F: {partialF:.2f})')
        print(f' {" ".join(sorted(best_vars))}')
        model = best_model
        include = best_vars
    return include, model



import gzip
from pathlib import Path

DEMO_FILE_DIR = Path(__file__).parent


def _readFile(filename):
    with gzip.open(filename) as f:
        return f.read().decode('utf-8')


def globalWarmingBlogs():
    return {
        # https://www.tomorrow.io/weather/blog/global-warming-status/
        'blog-1': _readFile(DEMO_FILE_DIR / 'global-warming-blog-1.txt.gz'),
        # https://blogs.scientificamerican.com/plugged-in/global-warming-explained-in-under-a-minute/
        'blog-2': _readFile(DEMO_FILE_DIR / 'global-warming-blog-2.txt.gz'),
    }


def covid19Blogs():
    """ Three covid-19 related blog posts from Eduardo Levy Yeyati """
    return {
        # https://www.brookings.edu/research/social-and-economic-impact-of-covid-19/
        'blog-1': _readFile(DEMO_FILE_DIR / 'covid-19-blog-1.txt.gz'),
        # https://www.brookings.edu/opinions/the-covid-reset-latin-america-needs/
        'blog-2': _readFile(DEMO_FILE_DIR / 'covid-19-blog-2.txt.gz'),
        # https://voxeu.org/article/covid-19-americas-north-south-differences-and-labour-market-channel
        'blog-3': _readFile(DEMO_FILE_DIR / 'covid-19-blog-3.txt.gz'),
    }



# mypy: disallow_untyped_defs,disallow_untyped_calls
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from dataclasses import dataclass
from typing import Dict, cast, List, Union
import numpy as np
from scipy import stats


@dataclass
class Distribution:
    pass


@dataclass
class BetaDistribution(Distribution):
    # shape parameters
    a: int
    b: int

    def theta(self):
        return self.a / (self.a + self.b)


@dataclass
class GammaDistribution(Distribution):
    shape: float
    rate: float  # scale is 1/rate


@ dataclass
class Mixture:
    probabilities: List[float]
    distributions: List[Distribution]


def updateBetaMixture(mixture: Mixture, data: List[int]) -> Mixture:
    """ Update of a mixture of beta distribution given data

    mixture: mixture of beta distribution characterized by
        probabilities: list of probabilities of the beta components of the prior
        distributions: list of beta distribution of the prior
    data: list with number of successes and number of failures

    returns: update mixture of beta distributions
    """
    distributions = cast(List[BetaDistribution], mixture.distributions)
    probabilities = np.array(mixture.probabilities)

    # Add successes and failures to the shape parameter of the beta distribution
    s, f = data
    postBeta = [BetaDistribution(a=bd.a + s, b=bd.b + f)
                for bd in distributions]
    p = np.array([bd.a / (bd.a + bd.b) for bd in postBeta])

    mProb = np.zeros(len(p))
    for i, (pi, bd, postBd) in enumerate(zip(p, distributions, postBeta)):
        mProb[i] = np.exp(stats.binom.logpmf(s, s+f, pi) +
                          stats.beta(a=bd.a, b=bd.b).logpdf(pi) -
                          stats.beta(a=postBd.a, b=postBd.b).logpdf(pi))
    # update the probabilities of the prior
    postProbs = (probabilities * mProb) / np.sum(probabilities * mProb)
    return Mixture(probabilities=list(postProbs), distributions=list(postBeta))


def updateGammaMixture(mixture: Mixture, data: Dict[str, List[float]]) -> Mixture:
    """ Update of a mixture of gamma distribution given data

    mixture: mixture of gamma distribution characterized by
        probabilities: list of probabilities of the gamma components of the prior
        distributions: list of gamma distribution of the prior
    data: dictionary of lists, 'y' = counts, 't' = time intervals

    returns: updated mixture of gamma distributions
    """
    distributions = cast(List[GammaDistribution], mixture.distributions)
    probabilities = np.array(mixture.probabilities)

    y = data['y']
    t = data['t']
    ysum = np.sum(y)
    tsum = np.sum(t)
    postGamma = [GammaDistribution(shape=d.shape + ysum, rate=d.rate+tsum)
                 for d in distributions]
    L = np.array([d.shape / d.rate for d in postGamma])

    loglike = np.zeros(len(L))
    for yi, ti in zip(y, t):
        loglike += stats.poisson.logpmf(yi, L * ti)

    mProb = np.zeros(len(L))
    for i, (Li, bd, postBd) in enumerate(zip(L, distributions, postGamma)):
        mProb[i] = np.exp(loglike[i] +
                          stats.gamma(bd.shape, scale=1/bd.rate).logpdf(Li) -
                          stats.gamma(postBd.shape, scale=1/postBd.rate).logpdf(Li))
    # update the probabilities of the prior
    postProbs = (probabilities * mProb) / np.sum(probabilities * mProb)
    return Mixture(probabilities=list(postProbs), distributions=list(postGamma))



import unittest

from mistat.design.doeUtilities import (_reduceTreatment, aliasesInSubgroup,
                                        subgroupOfDefining)


class TestDoeUtilities(unittest.TestCase):
    def test_subgroupOfDefining(self):
        expected = ['']
        assert subgroupOfDefining(['']) == sorted(expected)
        expected = ['', 'ABCH', 'ABEFG', 'CEFGH']
        assert subgroupOfDefining(['ABCH', 'ABEFG']) == sorted(expected)
        expected = [*expected, 'BDEFH', 'ACDEF', 'ADGH', 'BCDG']
        assert subgroupOfDefining(['ABCH', 'ABEFG', 'BDEFH']) == sorted(expected)

        expected = ['(1)', 'ABCH', 'ABEFG', 'CEFGH']
        assert subgroupOfDefining(['ABCH', 'ABEFG'], noTreatment='(1)') == sorted(expected)

    def test_reduceTreatment(self):
        assert _reduceTreatment('') == ''
        assert _reduceTreatment('A') == 'A'
        assert _reduceTreatment('AB') == 'AB'
        assert _reduceTreatment('ABA') == 'B'
        assert _reduceTreatment('BAAA') == 'AB'
        assert _reduceTreatment('BAAACA') == 'BC'

    def test_aliasesInSubgroup(self):
        expected = sorted({'BCH', 'BEFG', 'ACEFGH', 'ABDEFH', 'CDEF', 'DGH', 'ABCDG',
                           'ABCEFH', 'EF', 'CGH', 'ABG', 'ACD', 'BDH', 'BCDEFG', 'ADEFGH'})
        assert aliasesInSubgroup('A', ['ABCH', 'ABEFG', 'BDEFH', 'BCEFH']) == expected

        assert aliasesInSubgroup('A', ['AB']) == ['B']



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest
from pathlib import Path

import pandas as pd
import pytest

import mistat
from mistat.data import DATA_DIR, get_description_file


class TestData(unittest.TestCase):
    def test_load_data(self):
        with pytest.raises(ValueError):
            mistat.load_data('unknown data file')

        for name in ('CYCLT.csv.gz', 'CYCLT.csv', 'CYCLT'):
            data = mistat.load_data(name)
            assert isinstance(data, pd.Series)

    def test_load_data_all(self):
        for name in Path(DATA_DIR).glob('*.csv.gz'):
            data = mistat.load_data(name.name)
            assert isinstance(data, (pd.Series, pd.DataFrame))
            assert len(data.shape) <= 2
            if len(data.shape) == 1:
                assert isinstance(data, pd.Series)
            else:
                assert isinstance(data, pd.DataFrame)
                assert data.shape[1] > 1

    def _test_data_description_exists(self):
        for name in Path(DATA_DIR).glob('*.Rd'):
            data_file = name.with_suffix('.csv.gz')
            assert data_file.exists()
        for name in Path(DATA_DIR).glob('*.csv.gz'):
            description_file = get_description_file(name.name)
            assert description_file.exists()

    def _test_describe_data(self):
        with pytest.raises(ValueError):
            mistat.describe_data('unknown data file')

        for name in ('CYCLT.csv.gz', 'CYCLT.csv', 'CYCLT'):
            text = mistat.describe_data(name)
            assert isinstance(text, str)
            assert 'Description' in text

        for name in Path(DATA_DIR).glob('*.csv.gz'):
            text = mistat.describe_data(name.name)
            assert isinstance(text, str)
            assert 'Description' in text

    def _test_prepare_Rd_HTML(self):
        for name in Path(DATA_DIR / 'Rd').glob('*.Rd'):
            name = name.with_suffix('').name
            print(f'Rd2HTML(file("{name}.Rd"), "{name}.html")')
        for name in Path(DATA_DIR / 'Rd').glob('*.Rd'):
            name = name.with_suffix('').name
            print(f'pandoc {name}.html -t rst -o {name}.rst')
        for name in Path(DATA_DIR / 'Rd').glob('*.Rd'):
            name = name.with_suffix('').name
            print(f'pandoc {name}.rst -t markdown -o {name}.md')

    def test_specialDatasets(self):
        data = mistat.load_data('PROCESS_SEGMENT')
        assert len(data['X']) == 1897
        assert len(data['Z']) == 1002



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest

from mistat.data import load_data
from mistat.mqcc.multivariateQualityControlChart import \
    MultivariateQualityControlChart


class TestMultivariateQualityControlChart(unittest.TestCase):
    def test_T2single_chart(self):
        almpin = load_data('ALMPIN')
        base = almpin.iloc[:30, ]
        almpinSubset = almpin.iloc[30:, ]

        center = base.mean()
        cov = base.cov()

        _ = MultivariateQualityControlChart(almpinSubset, qcc_type='T2single',
                                            center=center, cov=cov)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest
from pathlib import Path

import numpy as np
import pandas as pd

from mistat.qcc.qualityControlChart import QualityControlChart, qcc_groups
from mistat.qcc.rules import run_length_encoding


class TestRules(unittest.TestCase):
    def test_shewhartRules(self):
        #         qcc = QualityControlChart(load_data('JANDEFECT'), qcc_type='np', sizes=100)
        #         print(qcc.violations)

        out = [idx - 1 for idx in [9, 10, 30, 35, 45, 64, 65, 74, 75, 85, 99, 100]]
        pistonrings = pd.read_csv(Path(__file__).parent / 'data' / 'pistonrings.csv')
        pistonrings = pistonrings.drop(pistonrings.index[out])
        diameter = qcc_groups(pistonrings['diameter'], pistonrings['sample'])
        _ = QualityControlChart(diameter, qcc_type='xbar', center=74, std_dev=0.0098)

    def test_run_length_encoding(self):
        assert run_length_encoding('abbccc') == [(1, 'a'), (2, 'b'), (3, 'c')]
        assert run_length_encoding('abbcccaaaaa') == [(1, 'a'), (2, 'b'), (3, 'c'), (5, 'a')]
        assert run_length_encoding([]) == []
        np.testing.assert_array_equal(run_length_encoding('abbccc', as_list=True), [1, 2, 2, 3, 3, 3])



""" Accompanying source code for

Ron Kenett, Shelemyahu Zacks, Peter Gedeck
- Modern Statistics: A Computer Based Approach with Python
- Industrial Statistics: A Computer Based Approach with Python
"""
from .data import describe_data, load_data
from .design import (FDS_Plot, addTreatments, calculateInteractions,
                     calculateMainEffects, getModelMatrix, interactionPlot,
                     mainEffectsPlot)
from .ecdf import plotECDF
from .ml import plot_dendrogram
from .mqcc import MahalanobisT2, MultivariateQualityControlChart
from .qcc import (EWMA, Cusum, ParetoChart, ProcessCapability,
                  QualityControlChart, cusumArl, cusumPfaCed, qcc_groups,
                  qccStatistics, shroArlPfaCedNorm)
from .randomizationTest import randomizationTest
from .regression import stepwise_regression
from .reliability import availabilityEBD, renewalEBD
from .rsm import ResponseSurfaceMethod
from .runsTest import runsTest
from .simulation import (PistonSimulator, PowerCircuitSimulation,
                         SimulationResult, simulationGroup)
from .stem_leaf import stemLeafDiagram
from .timeseries import (dlmLinearGrowth, masPredictor, normRandomWalk,
                         optimalLinearPredictor, predictARMA,
                         quadraticPredictor, simulateARMA)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from collections import namedtuple

import numpy as np
import pandas as pd
from scipy import linalg, stats

from mistat.qcc.statistics import QCCStatistics

GroupStatistics = namedtuple('GroupStatistics', 'statistics,means,center,cov')
DataSizes = namedtuple('DataSizes', 'num_samples,samples_sizes,num_variables')


class MQCCStatistics(QCCStatistics):

    def __init__(self):  # pylint: disable=super-init-not-called
        self.default = None
        self.statistics = {}
        self.register(T2_statistic, default=True)
        self.register(T2single_statistic)


class T2_statistic:
    """ Statistics used in computing and drawing a Shewhart xbar chart """
    qcc_type = 't2'
    description = ('T2', 'Hotelling T^2 chart for subgrouped data')

    def get_sizes(self, data):
        shapes = {k: v.shape for k, v in data.items()}
        num_samples = [v[0] for v in shapes.values()]  # m
        if len(set(num_samples)) == 1:
            num_samples = num_samples[0]
        sample_sizes = [v[1] for v in shapes.values()]  # n
        if len(set(sample_sizes)) == 1:
            sample_sizes = sample_sizes[0]
        p = len(data)  # number of variables
        return DataSizes(num_samples, sample_sizes, p)

    def stats(self, data, center=None, cov=None):
        num_samples, sample_sizes, p = self.get_sizes(data)

        # within sample means
        means = pd.DataFrame({k: np.nanmean(v, axis=1) for k, v in data.items()})
        # overall mean
        if center is None:
            center = pd.DataFrame([{k: np.nanmean(v) for k, v in data.items()}])
        else:
            center = pd.DataFrame(center)
        x = means.sub(center.values, axis=1)

        if cov is None:
            cov = np.zeros([p, p])
            for i in range(num_samples):
                sampleMean = means.iloc[i, :].values
                d = np.transpose(np.array([v[i] for k, v in data.items()]))
                d = d - sampleMean
                d = np.transpose(d) @ d
                cov += d / (sample_sizes - 1)
            cov = cov / num_samples
        cov_inv = linalg.inv(cov)

        T2 = sample_sizes * x.apply(lambda x: x @ cov_inv @ x, axis=1)
        return GroupStatistics(statistics=T2, means=means, center=center, cov=cov)

    def limits(self, ngroups, size, nvars, conf):
        m = ngroups  # num. of samples
        n = size  # samples sizes
        p = nvars  # num. of variables

        upper = p * (n - 1) / (m * n - m - p + 1) * stats.f(p, m * n - m - p + 1).ppf(conf)

        # Phase 1 control limits
        ucl = (m - 1) * upper
        lcl = 0
        control = pd.DataFrame([{'LCL': lcl, 'UCL': ucl}])

        # Phase 2 prediction limits
        ucl = (m + 1) * upper
        lcl = 0
        prediction = pd.DataFrame([{'LPL': lcl, 'UPL': ucl}])

        return {'control': control, 'prediction': prediction}


class T2single_statistic:
    """ Statistics used in computing and drawing a Shewhart xbar chart """
    qcc_type = 't2single'
    description = ('T2 single', 'Hotelling T^2 chart for individual observations')

    def get_sizes(self, data):
        return DataSizes(data.shape[0], 1, data.shape[1])

    def stats(self, data, center=None, cov=None):
        data = pd.DataFrame(data)
        m, _, _ = self.get_sizes(data)

        if center is None:
            center = np.mean(data, axis=0)
        else:
            center = pd.DataFrame(center)
        x = data.sub(center.values.flatten(), axis=1)
        if cov is None:
            cov = np.transpose(x) @ x
            cov = cov / (m - 1)
        cov_inv = linalg.inv(cov)
        T2 = x.apply(lambda x: x @ cov_inv @ x, axis=1)
        return GroupStatistics(statistics=T2, means=data, center=center, cov=cov)

    def limits(self, ngroups, size, nvars, conf):
        m = ngroups  # num. of samples
        # n = size  # samples sizes # @UnusedVariable
        p = nvars  # num. of variables

        # Phase 1 control limits
        ucl = (m - 1)**2 / m * stats.beta(p / 2, (m - p - 1) / 2).ppf(conf)
        lcl = 0
        control = pd.DataFrame([{'LCL': lcl, 'UCL': ucl}])

        # Phase 2 prediction limits
        ucl = p * (m + 1) * (m - 1) / (m * (m - p)) * stats.f(p, m - p).ppf(conf)
        lcl = 0
        prediction = pd.DataFrame([{'LPL': lcl, 'UPL': ucl}])
        return {'control': control, 'prediction': prediction}


# > limits.T2.single
# function (ngroups, size = 1, nvars, conf)
# {
#     m <- ngroups
#     n <- size
#     p <- nvars
#     ucl <- (m - 1)^2/m * qbeta(conf, p/2, (m - p - 1)/2)
#     lcl <- 0
#     ctrl.limits <- matrix(c(lcl, ucl), ncol = 2)
#     ucl <- p * (m + 1) * (m - 1)/(m * (m - p)) * qf(conf, p,
#         m - p)
#     lcl <- 0
#     pred.limits <- matrix(c(lcl, ucl), ncol = 2)
#     rownames(ctrl.limits) <- rownames(pred.limits) <- rep("",
#         nrow(pred.limits))
#     colnames(ctrl.limits) <- c("LCL", "UCL")
#     colnames(pred.limits) <- c("LPL", "UPL")
#     return(list(control = ctrl.limits, prediction = pred.limits))
# }

mqccStatistics = MQCCStatistics()



# pylint: disable=too-many-arguments
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from itertools import combinations, product

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import patsy
import seaborn as sns
from scipy import stats


def calculateMainEffects(df, response, factors=None):
    if response not in df.columns:
        raise ValueError(f'Response {response} not found in data')
    factors = factors or [c for c in df.columns if c != response]
    mainEffects = []
    for column in factors:
        for level, g in sorted(df[response].groupby(df[column])):
            mainEffects.append({'factor': column, 'level': level, 'mean': g.mean()})
    return pd.DataFrame(mainEffects)


def calculateInteractions(df, response, factors=None):
    factors = factors or [c for c in df.columns if c != response]
    interactions = []
    for f1, f2 in product(factors, factors):
        for (l1, l2), g in df[response].groupby([df[f1], df[f2]]):
            interactions.append({'f1': f1, 'l1': l1, 'f2': f2, 'l2': l2, 'mean': g.mean()})
    return pd.DataFrame(interactions)


def mainEffectsPlot(df, response, factors=None, col_wrap=None, aspect=0.4, height=3, continuous=False):
    factors = factors or [c for c in df.columns if c != response]
    mainEffects = calculateMainEffects(df, response, factors)
    col_wrap = col_wrap or len(factors)

    def effectPlot(x, y, **kwargs):
        ax = plt.gca()
        ax.plot(x, y, marker='s')
        ax.axhline(kwargs['meanEffect'], color='grey', linestyle=':')
        x = sorted(x)
        d = 0.35 * (x[-1] - x[0])
        ax.set_xlim(x[0] - d, x[-1] + d)
        if not continuous:
            ax.set_xticks(x)

    dy = 0.15 * (mainEffects['mean'].max() - mainEffects['mean'].min())
    ylim = [mainEffects['mean'].min() - dy, mainEffects['mean'].max() + dy]
    meanEffect = df[response].mean()

    g = sns.FacetGrid(mainEffects, col='factor', col_wrap=col_wrap, height=height,
                      sharex=False, aspect=aspect, despine=True, ylim=ylim)
    g.map(effectPlot, 'level', 'mean', meanEffect=meanEffect).set_titles('')
    g.fig.subplots_adjust(wspace=.02)
    for idx, column in enumerate(factors):
        g.axes[idx].set_xlabel(column)


def interactionPlot(df, response, factors=None):
    factors = factors or [c for c in df.columns if c != response]
    interactions = calculateInteractions(df, response, factors)

    def setLimits(set_lim, values, f):
        v = sorted(set(values))
        d = f * (v[-1] - v[0])
        set_lim(v[0] - d, v[-1] + d)

    def effectPlot(x1, x2, y, **kwargs):
        x2values = sorted(set(x2), reverse=True)
        linestyle = ['-', '--', ':']
        color = ['black', 'red', 'blue']
        marker = ['s', '^', 'o']

        ax = plt.gca()
        setLimits(ax.set_xlim, x1, 0.35)
        ax.set_xticks(sorted(set(x1)))
        if len(x1) == len(set(x1)):
            for idx, x2i in enumerate(x2values):
                yi = (idx + 1) / (len(x2values) + 1)
                i = idx % len(linestyle)
                ax.text(0.2, yi, f'{x2i:.4g}', transform=ax.transAxes, verticalalignment='center')
                ax.plot(0.1, yi, marker=marker[i], color=color[i], linestyle=linestyle[i], transform=ax.transAxes)
            return
        for x2i, group in pd.DataFrame({'x1': x1, 'x2': x2, 'y': y}).groupby('x2'):
            i = x2values.index(x2i) % len(linestyle)
            group.plot.line(x='x1', y='y', ax=ax,
                            marker=marker[i], color=color[i], linestyle=linestyle[i])

    dy = 0.15 * (interactions['mean'].max() - interactions['mean'].min())
    ylim = [interactions['mean'].min() - dy, interactions['mean'].max() + dy]

    g = sns.FacetGrid(interactions, col='f1', row='f2', sharex=False, despine=True, ylim=ylim, height=1.5)
    g.map(effectPlot, 'l1', 'l2', 'mean').set_titles('')
    g.fig.subplots_adjust(wspace=.05, hspace=.05)
    for idx, column in enumerate(factors):
        g.axes[-1][idx].set_xlabel(column)
        g.axes[idx][0].set_ylabel('')
        ax = g.axes[idx][idx]
        ax.text(0.9, 0.5, column, transform=ax.transAxes, verticalalignment='center', horizontalalignment='right')
    return ax


def getModelMatrix(design, mod=0, maxscale=1):
    ''' Convert design to model matrix. Will rescale design to code values '''
    design = design - design.mean()
    design = maxscale * design / design.max()

    kvar = design.shape[1]
    names = [f'x{i + 1}' for i in range(kvar)]
    design.columns = names

    formula = list(names)
    if mod >= 1:
        formula.extend(f'{f1}:{f2}' for f1, f2 in combinations(names, 2))
    if mod == 2:
        formula.extend(f'np.power({f1}, 2)' for f1 in names)

    dm = patsy.dmatrix(f'~ {"+".join(formula)}', data=design, return_type='dataframe')  # @UndefinedVariable

    def cleanColumnName(c):
        if not c.startswith('np.power'):
            return c
        c = c.split('(')[1].split(',')[0]
        return f'{c}**2'
    dm.columns = [cleanColumnName(c) for c in dm.columns]
    return dm


def FDS_Plot(design, mod=0, ax=None, plotkw=None, label='y', maxscale=1):
    ''' Fraction of design space plot '''
    dm = getModelMatrix(design, mod=mod, maxscale=maxscale)

    terms = dm.columns
    kvar = design.shape[1]
    names = [f'x{i + 1}' for i in range(kvar)]

    XtX = np.matmul(np.transpose(dm).values, dm.values)
    XtXI = pd.DataFrame(np.linalg.inv(XtX),
                        index=terms, columns=terms)

    fX = {'Int': [1] * 5000}
    fX.update({f: stats.uniform.rvs(loc=-1, scale=2, size=5000) for f in names})
    if mod >= 1:
        for f1, f2 in combinations(names, 2):
            fX[f'{f1}:{f2}'] = fX[f1] * fX[f2]
    if mod == 2:
        for f in names:
            fX[f'np.power({f}, 2)'] = fX[f] * fX[f]
    fX = pd.DataFrame(fX)
    p1 = np.matmul(fX.values, XtXI)
    v = np.diag(np.matmul(p1.values, np.transpose(fX.values)))
    vi = np.sort(v)

    # calculate fraction of design space
    fs = np.arange(1, len(vi) + 1) / len(vi)
    plotkw = plotkw or {}
    plotkw['color'] = plotkw.get('color', 'black')
    # plotkw['legend'] = False
    ax = pd.DataFrame({'x': fs, label: vi}).plot(x='x', y=label, ax=ax, **plotkw)
    ax.axvline(0.5, linestyle=':', color='grey')
    ax.axhline((vi[2499] + vi[2500]) / 2, linestyle=':', color='grey')
    ax.set_xlabel('Fraction of Space')
    ax.set_ylabel('Relative Prediction Variance')
    return ax



# mypy: disallow_untyped_defs,disallow_untyped_calls
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from typing import List, Union

import numpy as np
import pandas as pd
from scipy import stats

from .dodge_base import AcceptanceSamplingPlan
# ' Lot Sensitive Compliance Sampling Plans
# '
# ' The lot sensitive compliance sampling plans for given parameters.
# '
# '
# ' @param N the lot size
# ' @param LTPD the lot tolerance percent defective, also known as the limiting
# ' quality
# ' @param beta consumer risk
# ' @param p fraction nonconforming
# ' @param Plots logical indicating if the four plots are required
# ' @author Raj Govindaraju with minor editing by Jonathan Godfrey
# ' @references Schilling, E.G. (1978) \dQuote{A Lot Sensitive Sampling Plan for
# ' Compliance Testing and Acceptance Inspection}, \emph{Journal of Quality
# ' Technology} \bold{10}(2), pp47-51.


def lotSensitiveComplianceSampPlan(N: int, LTPD: float, beta: float,
                                   p: Union[List[float], np.ndarray] = None) -> AcceptanceSamplingPlan:
    if p is None:
        p = np.arange(0, 0.301, 0.001)
    p = np.array(p)

    f = 1 - (beta**(1 / (LTPD * N)))
    n = round(f * N)
    OC = (1 - f)**(N * p)
    AOQ = (N - n) * p * OC / N
    ATI = n * OC + N * (1 - OC)
    return AcceptanceSamplingPlan(p=list(p), OC=list(OC), n=[n] * len(p), AOQ=list(AOQ),
                                  ATI=list(ATI), ASN=None)


# ' Variable Sampling Plans
# '
# ' Variable sampling plans for known and unknown sigma, evaluated for given
# ' parameters.
# '
# '
# ' @aliases VSPKnown VSPUnknown
# ' @param N the lot size
# ' @param n the sample size
# ' @param k the acceptability constant
# ' @param Pa fraction nonconforming
# ' @param Plots logical indicating whether the four plots are required
# ' @author Raj Govindaraju with minor editing by Jonathan Godfrey
# ' @keywords ~kwd1 ~kwd2
# ' @examples
# '
# ' VSPKnown(1000, 20,1)
# ' VSPUnknown(1000, 20,1)
# '

def variableSampPlanKnown(N: int, n: int, k: float,
                          pa: Union[List[float], np.ndarray] = None) -> AcceptanceSamplingPlan:
    if pa is None:
        pa = np.arange(0, 1.001, 0.001)
    pa = np.array(pa)

    zpa = stats.norm.ppf(pa)
    zp = k + zpa / np.sqrt(n)
    p = 1 - stats.norm.cdf(zp)
    OC = pa
    AOQ = (N - n) * p * OC / N
    ATI = n * OC + N * (1 - OC)
    return AcceptanceSamplingPlan(p=p, OC=OC, n=[n] * len(p), AOQ=AOQ, ATI=ATI, ASN=None)  # type: ignore


def variableSampPlanUnknown(N: int, n: int, k: float,
                            pa: Union[List[float], np.ndarray] = None) -> AcceptanceSamplingPlan:
    if pa is None:
        pa = np.arange(0, 1.001, 0.001)
    pa = np.array(pa)

    zpa = stats.norm.ppf(pa)
    k1 = np.sqrt(1 + k * k / 2)
    zp = k + k1 * zpa / np.sqrt(n)
    p = 1 - stats.norm.cdf(zp)
    OC = pa
    AOQ = (N - n) * p * OC / N
    ATI = n * OC + N * (1 - OC)
    return AcceptanceSamplingPlan(p=p, OC=OC, n=[n] * len(p), AOQ=AOQ, ATI=ATI, ASN=None)  # type: ignore

# ' Variable Sampling Plan Design
# '
# ' Design the variable sampling plan for given AQL, alpha, LQL, and beta.
# '
# '
# ' @param AQL Acceptable quality level
# ' @param alpha producer's risk
# ' @param LQL Limiting quality level
# ' @param beta consumers' risk
# ' @author Raj Govindaraju with minor editing by Jonathan Godfrey
# ' @keywords ~kwd1 ~kwd2
# ' @examples
# '
# ' VSPDesign(AQL=0.01, alpha=0.05, LQL=0.04, beta=0.05)


def VSPDesign(AQL: float, alpha: float, LQL: float, beta: float) -> pd.Series:
    zp1 = stats.norm.ppf(1-AQL)
    zp2 = stats.norm.ppf(1-LQL)
    zpa1 = stats.norm.ppf(1-alpha)
    zpa2 = stats.norm.ppf(1-beta)
    k = (zp2 * zpa1 + zp1 * zpa2) / (zpa1 + zpa2)
    n = (zpa1 + zpa2) / (zp1 - zp2)
    n = n * n
    n = round(n)
    n_unknown = n * (1 + k * k / 2)
    return pd.Series({
        'k': k,
        'n': n,
        'n_unknown': n_unknown,
    })
